#include <opcerr.h>
#include <opcdebug.h>
#include <opcassert.h>

#include <opcdatadestructor.h>

static inline void opcDataDestructorZeroContext(OPCDataDestructorContext ctx) {
    assert(ctx);
    
    ctx->chunk = NULL;
    ctx->lock = NULL;
}

static void opcDataDestructorFreeContext(OPCDataDestructorContext ctx) {
    assert(ctx);
    if (ctx->chunk) opcFree(ctx->malloc, ctx->chunk);
    if (ctx->lock) LockDestroyContext(ctx->lock);
    opcDataDestructorZeroContext(ctx);
}

static void opcDataDestructorDestroyContext(OPCDataDestructorContext ctx) {
    opcDataDestructorFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

OPCDataDestructorContext opcDataDestructorCreateContext(OPCMallocContext mctx) {
    int err;
    OPCDataDestructorContext ctx;
    
    ctx = (OPCDataDestructorContext)opcMalloc(mctx,sizeof(struct OPCDataDestructorContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcDataDestructorZeroContext(ctx);
    } 
    return ctx;
}

int opcDataDestructorInitContext(OPCDataDestructorContext ctx, unsigned short int representations) {
    int err;
    long i;
    OPCDataChunkDestructorContext chunk;

    assert(ctx);
    assert(representations);
    
    opcDataDestructorZeroContext(ctx);
    
    chunk = (OPCDataChunkDestructorContext)opcMalloc(ctx->malloc, representations*sizeof(struct OPCDataChunkDestructorContextT));
    if (!chunk) return ERR_OPC_MEM;
    
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcDataDestructorFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	opcDataDestructorFreeContext(ctx);
	return err;
    }

    ctx->representations = representations;
    ctx->refs = 1;

    for (i=representations-1;i>=0;i--) {
	chunk[i].type = OPC_DATA_CHUNK_DESTRUCTOR_FREE;
	chunk[i].allocator = NULL;
    }
    
    return 0;
}

void opcDataDestructorRef(OPCDataDestructorContext ctx) {
    assert(ctx);

    LockWrite(ctx->lock);
    ctx->refs++;
    LockReturn(ctx->lock);
}

void opcDataDestructorUnref(OPCDataDestructorContext ctx) {
    assert(ctx);

    LockWrite(ctx->lock);
    if (!(--(ctx->refs))) {
	LockReturn(ctx->lock);
	opcDataDestructorDestroyContext(ctx);
    }
    LockReturn(ctx->lock);
}

int opcDataDestructorResize(OPCDataDestructorContext ctx, unsigned short int representations) {
    OPCDataChunkDestructorContext chunk;

    if (ctx->representations<representations) {
	chunk=opcRealloc(ctx->malloc, ctx->chunk, representations*sizeof(struct OPCDataChunkDestructorContextT));
	if (!chunk) return ERR_OPC_MEM;
	ctx->chunk = chunk;
    } else if (ctx->representations>representations) {
	if (ctx->refs==1) {
	    chunk=opcRealloc(ctx->malloc, ctx->chunk, representations*sizeof(struct OPCDataChunkDestructorContextT));
	    if (!chunk) return ERR_OPC_MEM;
	    ctx->chunk = chunk;
	} else  return ERR_OPC_BUSY;
    } 
    return 0;
}

int opcDataDestructorSet(OPCDataDestructorContext ctx, unsigned long chunk_id, OPCDataChunkDestructorType type, void *destructor, void *allocator) {
    OPCDataChunkDestructorContext chunk;

    assert(ctx);
    if ((!ctx)||(chunk_id>=ctx->representations)) return ERR_OPC_PARAM;

    switch(type) {
	case OPC_DATA_CHUNK_DESTRUCTOR_RETURN:
	    if (!destructor) return ERR_OPC_PARAM;
	break;
	case OPC_DATA_CHUNK_DESTRUCTOR_DESTROY:
	    if (!destructor) return ERR_OPC_PARAM;
	break;
	case OPC_DATA_CHUNK_DESTRUCTOR_FREE:
	break;
	default:
	return ERR_OPC_PARAM;
    }

    chunk=&ctx->chunk[chunk_id];
    chunk->type = type;
    chunk->destructor = destructor;
    chunk->allocator = allocator;
    return 0;
}

OPCDataChunkDestructorContext opcDataDestructorGet(OPCDataDestructorContext ctx, unsigned long chunk_id) {
    if (!ctx) return NULL;
    if ((!ctx->chunk)||(chunk_id>=ctx->representations)) return NULL;
    return ctx->chunk+chunk_id;
}

void *opcDataDestructorGetDestructor(OPCDataDestructorContext ctx, unsigned long chunk_id) {
    if (!ctx) return ctx;
    if ((!ctx->chunk)||(chunk_id>=ctx->representations)) return NULL;
    return ctx->chunk[chunk_id].destructor;
}

void *opcDataDestructorGetAllocator(OPCDataDestructorContext ctx, unsigned long chunk_id) {
    if (!ctx) return ctx;
    if ((!ctx->chunk)||(chunk_id>=ctx->representations)) return NULL;
    return ctx->chunk[chunk_id].allocator;
}

void opcDataDestructorDestroyChunk(OPCDataDestructorContext ctx, OPCDataChunkContext chunk) {
    OPCDataChunkDestructorContext destructor;
    OPCRepresentationData data;
    long i;

    assert(ctx);
    assert(chunk);
    assert(chunk->supported_representations <= ctx->representations);
        
    data = chunk->data;
    destructor = ctx->chunk;
    for (i=chunk->supported_representations-1;i>=0;i--) {
        if (data[i].data) {
	    switch(destructor[i].type) {
		case OPC_DATA_CHUNK_DESTRUCTOR_RETURN:
		    ((OPCReturnFunction)(destructor[i].destructor))(((data[i].ctx)?data[i].ctx:destructor[i].allocator), data[i].data);
		break;
		case OPC_DATA_CHUNK_DESTRUCTOR_DESTROY:
		    ((OPCDestroyContextFunction)(destructor[i].destructor))(data[i].data);
		break;
		case OPC_DATA_CHUNK_DESTRUCTOR_FREE:
		    opcFree(((data[i].ctx)?data[i].ctx:destructor[i].allocator), data[i].data);
		break;
	    }
	}
    }
    opcDataDestructorUnref(ctx);
}
