#define _OPC_DATA_FAST_C

#include <opclib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <opcdatachunk.h>
#include <opcdatafast.h>

OPCDataFastOptions opcDataFastGetDefaultOptions() {
    int err;
    OPCDataFastOptions options;

    options = (OPCDataFastOptions)malloc(sizeof(struct OPCDataFastOptionsT));
    if (!options) return options;
    
    err = opcDataFastInitDefaultOptions(options);
    if (err) {
	opcDataFastDestroyOptions(options);
	return NULL;
    }
    return options;
}

int opcDataFastInitDefaultOptions(OPCDataFastOptions options) {
    ds_assert_param(options);
    memcpy(options, &opc_datamanager_fast_options, sizeof(struct OPCDataFastOptionsT));
    dsDataBufferInitDefaultOptions(&options->allocator);
    return 0;
}

int opcDataFastOptionsSetAcquireFunction(OPCDataFastOptions options, OPCDataManagerAcquireFunction acquire) {
    ds_assert_param(options);
    options->acquire = acquire;
    return 0;
}

int opcDataFastOptionsSetUserizeFunction(OPCDataFastOptions options, OPCRepresentationGetUserDataFunction get) {
    ds_assert_param(options);
    options->get = get;
    return 0;
}

int opcDataFastOptionsSetBufferSize(OPCDataFastOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->allocator.minimal_size = min;
    options->allocator.maximal_size = max;
    return 0;
}

int opcDataFastOptionsSetCacheSize(OPCDataFastOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->buffer_min = min;
    options->buffer_max = max;
    return 0;
}

void opcDataFastDestroyOptions(OPCDataFastOptions options) {
    if (options) free(options);
}


void opcDataFastZeroContext(OPCDataFastContext ctx) {
    ds_assert_return(ctx);

    ctx->allocator = NULL;
    ctx->chunks = NULL;
    ctx->current = NULL;
    ctx->lock = NULL;
    ctx->chunk_lock = NULL;
    ctx->allocator_lock = NULL;
    ctx->grabed_allocator = 0;
    ctx->nchunks = 0;
}

static void opcDataFastFreeContext(OPCDataFastContext ctx) {
    OPCDataChunkContext chunk, nchunk;

    dsDebugServer(DS_DEBUG_NOTICE, "opcDataFast(%p): FreeContext", ctx);    

    ds_assert_return(ctx);
    
    for (chunk=ctx->chunks;chunk;chunk=nchunk) {
        nchunk = chunk->next;
	opcDataChunkWaitUnref(chunk, 0);
        opcDataChunkDestroyContext(chunk);
    }

    if (ctx->grabed_allocator) opcDataFastReturnAllocator(ctx);
    if (ctx->allocator) dsDataBufferDestroyContext(ctx->allocator);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->chunk_lock) dsLockDestroyContext(ctx->chunk_lock);
    opcDataFastZeroContext(ctx);    
}

static void opcDataFastDestroyContext(OPCDataFastContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcDataFast(%p): DestroyContext", ctx);    
    opcDataFastFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


OPCDataFastContext opcDataFastCreateContext(DSMallocContext mctx) {
    OPCDataFastContext ctx;

    ctx = (OPCDataFastContext)dsMalloc(mctx,sizeof(struct OPCDataFastContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "opcDataFast(%p): CreateContext", ctx);    
    if (ctx) {
	ctx->malloc=mctx;
	ctx->refs = 1;
	opcDataFastZeroContext(ctx);
	memcpy(&ctx->manager, &opc_datamanager_fast, sizeof(struct OPCDataManagerContextT));
    }
    return ctx;
}


int opcDataFastInitContext(OPCDataFastContext ctx, OPCType type, OPCDataFastOptions options) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcDataFast(%p): InitContext (Type: %p)", ctx, type);

    ds_assert_param(ctx);
    ds_assert_param(type);
    ds_assert_param(options);
    ds_assert_options(options->chunk_pool);
    ds_assert_options(options->format);

    opcDataFastZeroContext(ctx);    

    ctx->type = type;
    ctx->chunk_pool = options->chunk_pool;
    ctx->format = options->format;
    ctx->repinfo.formats = options->formats;
    ctx->manager.Acquire = options->acquire;
    if (options->buffer_min) ctx->buffer_min = options->buffer_min;
    else ctx->buffer_min = 1;
    ctx->buffer_max = options->buffer_max;

    ctx->allocator = dsDataBufferCreateContext(ctx->malloc);
    if (!ctx->allocator) return DS_ERR_DATABUFFER_CREATE;
    
    err = dsDataBufferInitContext(ctx->allocator, &options->allocator);
    if (err) {
	opcDataFastFreeContext(ctx);
	return err;
    }

    opcRepresentationDataDataBuffer((OPCRepresentationInfo)&ctx->repinfo, ctx->allocator);
    if (options->get) ctx->repinfo.repinfo.GetUserData = options->get;
    ctx->repinfo.datafast = ctx;

    ctx->lock = dsLockCreateContext(ctx->malloc);
    ctx->chunk_lock = dsLockCreateContext(ctx->malloc);
    ctx->allocator_lock = dsLockCreateContext(ctx->malloc);
    if ((!ctx->lock)||(!ctx->chunk_lock)||(!ctx->allocator_lock)) {
	opcDataFastFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }

    err = dsLockInitContext(ctx->lock);
    if (!err) err = dsLockInitContext(ctx->chunk_lock);
    if (!err) err = dsLockInitContext(ctx->allocator_lock);
    if (err) {
	opcDataFastFreeContext(ctx);
	return err;
    }
    return 0;
}

int opcDataFastSetAcquireFunction(OPCDataFastContext ctx, OPCDataManagerAcquireFunction acquire) {
    ds_assert_param(ctx);
    
    ctx->manager.Acquire = acquire;
    return 0;
}

int opcDataFastSetUserizeFunction(OPCDataFastContext ctx, OPCRepresentationGetUserDataFunction get) {
    ds_assert_param(ctx);
    
    ctx->repinfo.repinfo.GetUserData = get;
    return 0;
}

int opcDataFastRef(OPCDataFastContext ctx) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->refs++;
    dsLockReturn(ctx->lock);
    return 0;
}

void opcDataFastUnref(OPCDataFastContext ctx) {
    ds_assert_return(ctx);

    if (!ctx->lock) {
	opcDataFastDestroyContext(ctx);
	return;
    }
    
    dsLockWrite(ctx->lock);
    if ((--ctx->refs)==0) {
	dsLockReturn(ctx->lock);
	opcDataFastDestroyContext(ctx);
	return;
    }
    dsLockReturn(ctx->lock);
}


static inline_function void opcDataFastClear(OPCDataFastContext ctx) {
    DSLockContext chunk_lock;
    OPCChunkPoolContext chunk_pool;
    OPCDataChunkContext chunk, nchunk;
    unsigned long buffer_min, buffer_max;
    unsigned long ref = 0;

	/* After chunk_lock locked number of referers can decrease, but not increase! */

    chunk_lock = ctx->chunk_lock;
    chunk_pool = ctx->chunk_pool;
    chunk = ctx->chunks;

    buffer_min = ctx->buffer_min;
    buffer_max = ctx->buffer_max;

    nchunk = chunk;

    if (buffer_max) 
        for (;((ctx->nchunks>buffer_max)&&(!ref));chunk=nchunk) {
            nchunk = chunk->next;
	    ctx->chunks=nchunk;
	    ctx->nchunks--;
	    dsLockReturn(chunk_lock);
	    opcChunkPoolReturnChunk(chunk_pool, chunk);
	    dsLockWrite(chunk_lock);
	    ref=opcDataChunkGetRefs(nchunk);
	}
	
    for (;((!ref)&&(ctx->nchunks>buffer_min));chunk=nchunk) {
	nchunk=chunk->next;
	ctx->chunks=nchunk;
	ctx->nchunks--;
	dsLockReturn(chunk_lock);
	opcChunkPoolReturnChunk(chunk_pool, chunk);
	dsLockWrite(chunk_lock);
	ref = opcDataChunkGetRefs(nchunk);
	if (!ref) ref = opcDataChunkGetDelayedRefs(nchunk);
    }
    
}


int opcDataFastPushData(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo info, const void *data, unsigned long size) {
    int err;
    OPCRepresentationData repdata;
    OPCRepresentationInfo repinfo;

    OPCDataChunkContext chunk;

    dsDebugServer(DS_DEBUG_DEBUG, "opcDataFast(%p): Push", ctx);

    ds_assert_param(ctx);
    ds_assert_param(data);


    if (ctx->grabed_allocator) return OPC_ERR_DATAFAST_ALLOCATOR_BUSY;
    
    repinfo = (OPCRepresentationInfo)&ctx->repinfo;
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, &repinfo, 1);
    if (!chunk) return OPC_ERR_CHUNKPOOL_GET;
    opcDataChunkCopyInfo(chunk, info);
    
    dsLockWrite(ctx->allocator_lock);
    repdata=opcDataChunkGetData(chunk, 0);
    err = opcFormatNativeImportData(ctx->format, ctx->type, (DSAllocatorContext)ctx->allocator, (OPCData)repdata, data, size);
	/* ToDo: prelock datachunk for all subscriptions may need it */
    dsLockReturn(ctx->allocator_lock);

    if (err) {
	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	return err;
    }
    
    dsLockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (dsTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current=chunk;

	if ((ctx->nchunks>ctx->buffer_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataFastClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    dsLockReturn(ctx->chunk_lock);
    
    return 0;
}

int opcDataFastPushAllocation(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo info, DSAllocation allocation) {
    OPCRepresentationInfo repinfo;
    OPCRepresentationData repdata;
    OPCDataChunkContext chunk;
    
    dsDebugServer(DS_DEBUG_DEBUG, "opcDataFast(%p): Push", ctx);

    ds_assert_param(ctx);
    ds_assert_param(allocation);

    repinfo = (OPCRepresentationInfo)&ctx->repinfo;
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, &repinfo, 1);
    if (!chunk) return OPC_ERR_CHUNKPOOL_GET;
    opcDataChunkCopyInfo(chunk, info);
    
    repdata=opcDataChunkGetData(chunk, 0);
    repdata->data = allocation;
    repdata->size = dsDataBufferSizeof(ctx->allocator, allocation, 1);
    repdata->type = OPC_DATA_TYPE_DATABUFFER;
	/* ToDo: prelock datachunk for all subscriptions may need it */

    dsLockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (dsTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current = chunk;

	if ((ctx->nchunks>ctx->buffer_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataFastClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    dsLockReturn(ctx->chunk_lock);

    
    return 0;
}

int opcDataFastGetData(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo *info, OPCRepresentationData *data) {
    OPCDataChunkContext current;

    ds_assert_complexctx(ctx, ctx->current);
    ds_assert_param(ctx->current);

    dsLockRead(ctx->chunk_lock);
    current = ctx->current;
    opcDataChunkRef(current);
    dsLockReturn(ctx->chunk_lock);

    if (data) *data=opcDataChunkGet(current, 0);
    if (info) *info=opcDataChunkGetInfo(current);
    
    opcDataFastRef(ctx);
    
    return 0;
}

void opcDataFastReturnData(OPCDataFastContext ctx, OPCDataInfo info) {
    ds_assert_return(ctx);
    if ((!ctx)||(!info)) return;

    dsLockWrite(ctx->chunk_lock);
    if ((!opcDataChunkUnrefByInfo(info))&&(ctx->chunks == opcDataChunkGetChunkByInfo(info)))
	opcDataFastClear(ctx);
    dsLockReturn(ctx->chunk_lock);
    
    opcDataFastUnref(ctx);
}


DSAllocatorContext opcDataFastGetAllocator(OPCDataFastContext ctx) {
    ds_assert_null(ctx);
    
    dsLockWrite(ctx->allocator_lock);
    ctx->grabed_allocator = 1;
    return (DSAllocatorContext)ctx->allocator;
}

void opcDataFastReturnAllocator(OPCDataFastContext ctx) {
    ds_assert_return(ctx);
    
    if (ctx->grabed_allocator) {
	ctx->grabed_allocator = 0;
	dsLockReturn(ctx->allocator_lock);
    }
}

void opcDataFastMantain(OPCDataFastContext ctx) {
    ds_assert_return(ctx);
    if (ctx->allocator) dsDataBufferMantain(ctx->allocator);
}
