#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include <dsmalloc.h>
#include <dslocale.h>
#include <dshashtable.h>
#include <dslock.h>
#include <dsstringlist.h>

#include <opctree.h>
//#include <opcproperty.h>
#include <opcitem.h>

void opcItemZeroDataContext(OPCItemContext ctx) {
    ds_assert_return(ctx);
    
    ctx->dm = NULL;
//    ctx->pctx=NULL;
}

void opcItemZeroContext(OPCItemContext ctx) {
    ds_assert_return(ctx);

    ctx->lock = NULL;
    ctx->client_cond = NULL;
    
    ctx->fullname=NULL;
    ctx->name=NULL;
    ctx->title=NULL;
    
    ctx->status=0;
    ctx->manual=2;
    ctx->mark=0;
    ctx->subitems=NULL;
    
    ctx->tree=NULL;
    
    ctx->vendor_filter=NULL;
    
    ctx->clients = 0;
    
    ctx->representation = 0;
    
    opcItemZeroDataContext(ctx);
}

OPCItemContext opcItemCreateContext(DSMallocContext mctx) {
    DSMallocContext mymalloc;
    OPCItemContext ctx;

    if (mctx) {
	mymalloc = dsMallocCreateContext(mctx);
	if (!mymalloc) return NULL;
    } else mymalloc = NULL;

    ctx = (OPCItemContext)dsMalloc(mymalloc, sizeof(struct OPCItemContextT));
    if (ctx) {
	ctx->malloc = mymalloc;
	opcItemZeroContext(ctx);
    } else dsMallocDestroyContext(mymalloc);

    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): CreateContext", ctx);

    return ctx;
}

int opcItemInitContext(OPCItemContext ctx, const char *fullname, const char *title) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): InitContext: \"%s\".", ctx, fullname);    
    
    ds_assert_param(ctx);
    ds_assert_param(fullname);
    
    opcItemZeroContext(ctx);
    
    err = dsExchangeString(ctx->malloc, &ctx->fullname, fullname);
    if (err) return DS_ERR_MEM;

/*
    ctx->name = strrchr(ctx->fullname, '/');
    if (ctx->name) ctx->name++;
    else ctx->name = ctx->fullname;
*/
    
    if (title) err=dsLocaleExchangeString(ctx->malloc,&ctx->title, title);
    else err=dsLocaleExchangeString(ctx->malloc,&ctx->title, fullname);
    
    if (err) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->client_cond = dsConditionCreateContext(ctx->malloc);
    if (!ctx->client_cond) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsConditionInitContext(ctx->client_cond);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->subitems = dsStringListCreate(ctx->malloc);
    if (!ctx->subitems) {
	opcItemFreeContext(ctx);
	return err;
    }

/*  This must be in over place and full destruction should occure only if haven't manual 
    sub(sub?)childs 
    
    ctx->pctx = opcPropertyGetInstance(sctx->pctx);
    if (!ctx->pctx) {
	opcItemFreeContext(ctx);
	return OPC_ERR_PROPERTIES_INSTANCE_INIT;
    }
*/	
    dsDebugServer(DS_DEBUG_INFO, "opcItem(%p): InitContext Ok", ctx);
    
    return 0;
}

void opcItemFreeDataContext(OPCItemContext ctx) {
    ds_assert_return(ctx);

    if (ctx->dm) opcDataManagerUnref(ctx->dm);
    opcItemZeroDataContext(ctx);
}

int opcItemFreeContext(OPCItemContext ctx) {
    int err = 0;
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): FreeContext %s", ctx, (ctx->fullname)?ctx->fullname:NULL);

    ds_assert_param(ctx);

    
    if ((ctx->tree)&&(opcTreeGetStatus(ctx->tree)!=OPC_TREE_DESTROYING_FLAG)) {
	if ((ctx->tree)&&(ctx->status&OPC_ITEM_LINKED_FLAG)&&(!opcItemIsMarked(ctx))) {
	    err=opcTreeUnregisterItem(ctx->tree, ctx);
	    if (err==OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) {
		opcItemResetManualMode(ctx);
	    } else {
		dsDebugServer(DS_DEBUG_CRITICAL, "opcItem(%p): Unregister failed: %s", ctx, dsError(err, NULL));
		err = 0;
	    }
	}
    }

    if (ctx->clients) {
	dsDebugServer(DS_DEBUG_DEBUG, "opcItem(%p): Waiting %lu clients to finish!", ctx, ctx->clients);
	
	dsConditionPreLock(ctx->client_cond);
	while (ctx->clients)
	    dsConditionWaitLocked(ctx->client_cond);
	dsConditionUnLock(ctx->client_cond);
	
	dsDebugServer(DS_DEBUG_DEBUG, "opcItem(%p): No more clients!", ctx);
    }

    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    opcItemFreeDataContext(ctx);
	// ToDo: Destroy properties

    if (err==OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) return err;


    if (ctx->subitems) dsStringListDestroy(ctx->subitems);

    // Unmark item
    if (ctx->client_cond) dsConditionDestroyContext(ctx->client_cond);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);

    if (ctx->fullname) dsExchangeString(ctx->malloc,&ctx->fullname,NULL);
    if (ctx->title) dsLocaleExchangeString(ctx->malloc,&ctx->title,NULL);

    opcItemZeroContext(ctx);

    dsDebugServer(DS_DEBUG_INFO, "opcItem(%p): FreeContext Ok", ctx);    
    return 0;
}

void opcItemDestroyContext(OPCItemContext ctx) {
    DSMallocContext mymalloc;
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): DestroyContext", ctx);    

    err=opcItemFreeContext(ctx);
    if (err!=OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) {
	mymalloc = ctx->malloc;
	dsFree(mymalloc,ctx);
	if (mymalloc) dsMallocDestroyContext(mymalloc);
    }
}

int opcItemLock(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    return dsLockWrite(ctx->lock);
}

int opcItemUnlock(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    return dsLockReturn(ctx->lock);
}

int opcItemGet(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->clients++;
    dsLockReturn(ctx->lock);

    dsConditionBroadcast(ctx->client_cond);

    return 0;
}

int opcItemReturn(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->clients--;
    dsLockReturn(ctx->lock);

    dsConditionBroadcast(ctx->client_cond);

    return 0;
}


int opcItemSetAssociation(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    return 0;
}

int opcItemSetManualMode(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->manual=1;
    dsLockReturn(ctx->lock);

    return 0;
}

int opcItemResetManualMode(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->manual=0;
    dsLockReturn(ctx->lock);

    return 0;
}

int opcItemSetVendorFilter(OPCItemContext ctx, OPCVendorFilter vendor_filter) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->vendor_filter = vendor_filter;
    dsLockReturn(ctx->lock);
    return 0;
}

int opcItemSetTitle(OPCItemContext ctx, const char *title) {
    ds_assert_param(ctx);
    ds_assert_param(title);
    dsLockWrite(ctx->lock);
    dsLocaleExchangeString(ctx->malloc,&ctx->title,title);
    dsLockReturn(ctx->lock);

    return 0;
}

dschar opcItemGetTitle(OPCItemContext ctx, const char *locale) {
    return dsLocaleGetText(ctx->title, locale);
}

static char *opcItemGetFullName_internal(OPCItemContext ctx) {
    char *res = NULL;
    ds_assert_null(ctx->fullname);
    dsExchangeString(ctx->malloc, &res, ctx->fullname);
    return res;
}

char *opcItemGetFullName(OPCItemContext ctx) {
    char *res;
    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);

    dsLockRead(ctx->lock);
    res = opcItemGetFullName_internal(ctx);
    dsLockReturn(ctx->lock);
    
    return res;
}

int opcItemBrowse(OPCItemContext ctx, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, DSStringList *result) {
    OPCItemContext ictx;
    DSStringList sl, res;
    char *fullname;
    int err;

    ds_assert_param(ctx);
    ds_assert_param(result);
    ds_assert_param(ctx->lock);
    
    *result = NULL;

    dsLockRead(ctx->lock);
    
    if ((ctx->subitems)&&(ctx->status&OPC_ITEM_LINKED_FLAG)) {
	res = dsStringListCreate(ctx->malloc?ctx->malloc->parrent:NULL);
	if (!res) err = DS_ERR_MEM;
	else err = 0;
    } else  err = OPC_ERR_SERVERTREE_ITEM_NOTLINKED;
    
    if (!err) {
	for (sl=ctx->subitems;sl;sl=sl->next) {
	    if (sl->item) {
		ictx = *(OPCItemContext*)sl->item;
		if (ictx) {
		    dsLockRead(ictx->lock);
		    if (((!browse_filter)||(!opcItemBrowseFilter(ictx, browse_filter)))&&
			((!name_filter)||(!opcItemNameFilter(ictx, name_filter)))&&
			((!vendor_filter)||((ctx->vendor_filter)&&(!ctx->vendor_filter(ictx, vendor_filter))))) {
			    fullname = ictx->fullname;
			    if (fullname) dsStringListAdd(res, fullname, (unsigned long)(strlen(fullname) + 1));
		    }
		    dsLockReturn(ictx->lock);
		}
	    }
	}
    }
    
    dsLockReturn(ctx->lock);
    
    if (err) *result=NULL;
    else *result=res;
    
    return err;
}

/*******************************************************************************
**************************** Non-Locking Functions  ****************************
*******************************************************************************/

int opcItemHaveChilds(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->status&OPC_ITEM_LINKED_FLAG) {
	ds_assert_param(ctx->subitems);
	return (!dsStringListIsEmpty(ctx->subitems));
    } 

    return 0;
}

int opcItemIsConnected(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    return ctx->status&OPC_ITEM_CONNECTED_FLAG;
}

/*******************************************************************************
*********************************** Filters ************************************
*******************************************************************************/
int opcItemBrowseFilter(OPCItemContext ctx, OPCItemBrowseFilter filter) {
    switch(filter) {
	case BRANCH:
	    if ((ctx->status&OPC_ITEM_CONNECTED_FLAG)&&(opcItemHaveChilds(ctx))) 
		return OPC_ERR_ITEM_FILTERED;
	case ITEM:
	    if ((ctx->status&OPC_ITEM_CONNECTED_FLAG)==0) 
		return OPC_ERR_ITEM_FILTERED;
    }
    return 0;
}


int opcItemNameFilter(OPCItemContext ctx, const char *filter) {
    return 0;
}


/*******************************************************************************
************************* Item SubItems Managment ******************************
*******************************************************************************/

int opcItemLink(OPCItemContext ctx, OPCItemContext ictx) {
    int err;
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    ds_assert_param(ictx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx->lock);
    
    dsLockWrite(ctx->lock);
    err=dsStringListAdd(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext));
    dsLockReturn(ctx->lock);
    return err;
}

int opcItemUnlink(OPCItemContext ctx, OPCItemContext ictx) {
    ds_assert_param(ctx);
    ds_assert_param(ictx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx->lock);

    dsLockWrite(ctx->lock);
    opcItemMarkParrentsCallback(ctx, ictx);
    opcTreeCleanMarked(ctx->tree);
    dsLockReturn(ctx->lock);
    return 0;
}
/*******************************************************************************
******************************* Tree Analysis **********************************
*******************************************************************************/

int opcItemHaveManualSubChilds(OPCItemContext ctx) {
    OPCItemContext ictx;
    DSStringList sl;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx = *(OPCItemContext*)sl->item;
	if (ictx) {
	    if (ictx->manual) return 1;
	    if (opcItemHaveManualSubChilds(ictx)) return 1;
	}
    }
    return 0;
}

int opcItemSetTree(OPCItemContext ctx, OPCTreeContext tree) {
    ds_assert_param(ctx);
    ds_assert_param(tree);
    ds_assert_param(ctx->tree==NULL);

    ctx->tree = tree;
    return 0;
}

/*******************************************************************************
************************* Deletion Marks Managment *****************************
*******************************************************************************/

void opcItemMark(OPCItemContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);
    
    if (!ctx->mark) {
	dsLockWrite(ctx->lock);
	ctx->mark=1;
    }
}

void opcItemMarkSubTree(OPCItemContext ctx) {
    DSStringList sl;
    OPCItemContext ictx;
    
    ds_assert_return(ctx);
    
    opcItemMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemMarkSubTree(ictx);
    }
}

void opcItemUnMark(OPCItemContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);
    
    if (ctx->mark) {
	ctx->mark=0;
	dsLockReturn(ctx->lock);
    }
}

void opcItemUnMarkSubTree(OPCItemContext ctx) {
    DSStringList sl;
    OPCItemContext ictx;

    ds_assert_return(ctx);
    
    opcItemUnMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemUnMarkSubTree(ictx);
    }
}

int opcItemIsMarked(OPCItemContext ctx) {
    ds_assert_param(ctx);
    return ctx->mark;
}

/*******************************************************************************
************************* Tree managment callbacks *****************************
*******************************************************************************/


int opcItemMarkChildsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    OPCItemContext ictx2;
    DSStringList sl;
    
    if ((ctx==ictx)||((opcItemIsMarked(ctx))&&(!ctx->manual))) return 0;
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx2 = *(OPCItemContext*)sl->item;
        if ((ictx2)&&(opcItemIsMarked(ictx2))) opcItemUnMarkSubTree(ictx2);
    }

    return 0;
}

int opcItemMarkParrentsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    ds_assert_param(ctx->tree);
    ds_assert_param(ictx);

    if (dsStringListRemove(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext*)))
	opcTreeMarkUnused(ctx->tree, ctx);
    return 0;
}

/*******************************************************************************
*************************** Properties and Data ********************************
*******************************************************************************/

OPCDataManagerContext opcItemRegisterDataManager(OPCItemContext ctx, OPCDataManagerContext dm) {
    ds_assert_null(ctx);
    
    dsLockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    ctx->dm = (OPCDataManagerContext)dm;
    dsLockReturn(ctx->lock);
    
    return dm;
}

OPCDataFastContext opcItemRegisterFastDataManager(OPCItemContext ctx, OPCType type, OPCDataFastOptions options, unsigned long flags) {
    int err;
    OPCDataFastContext dm;
    struct OPCDataFastOptionsT opts;
    OPCContext sctx = NULL;

    ds_assert_null(ctx);
    ds_assert_null(type);

    if ((!ctx)||(!type)) return NULL;
    
    if (!options) {
	opcDataFastInitDefaultOptions(&opts);
	options = &opts;
/*	options->flags = flags|OPC_DATA_FLAG_SEQUENTIAL; */
    }
    
    if (!options->formats) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->formats) return NULL;
	options->formats = sctx->formats;
    }
    if (!options->format) {
	if (!options->formats->native_format) return NULL;
	options->format = options->formats->native_format;
    }
    
    if (!options->chunk_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->chunks) return NULL;
	options->chunk_pool = sctx->chunks;
    }
    if (!options->allocator.pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->buffers) return NULL;
	options->allocator.pool = sctx->buffers;
    }
    
    if (type->size) options->allocator.fixed_allocation_size = type->size;
    
    dsLockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    
    dm = opcDataFastCreateContext(ctx->malloc);
    if (dm) {
	err = opcDataFastInitContext(dm, type, options);
	if (err) {
	    opcDataFastUnref(dm);
	    dm = NULL;
	} else {
	    ctx->dm = (OPCDataManagerContext)dm;
	    ctx->representation = OPC_REPRESENTATION_NATIVE;
	}
    }  else err = OPC_ERR_DATAFAST_CREATE;
    dsLockReturn(ctx->lock);
    
    if (err)
	dsDebugServer(DS_DEBUG_WARNING, "opcItem(%p): RegisterFastDataManager: %s", ctx, dsError(err, NULL));

    return dm;
}

OPCDataContext opcItemRegisterLegacyDataManager(OPCItemContext ctx, OPCType type, OPCDataOptions options, unsigned long flags) {
    int err;
    OPCDataContext dm;
    struct OPCDataOptionsT opts;
    OPCContext sctx = NULL;

    ds_assert_null(ctx);
    ds_assert_null(type);

    if ((!ctx)||(!type)) return NULL;
    
    if (!options) {
	opcDataInitDefaultOptions(&opts);
	options = &opts;
	options->flags = flags;
    }
    
    if (!options->formats) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->formats) return NULL;
	options->formats = sctx->formats;
    }
    if ((!options->formats->native_format)||(!options->formats->opcxml_format)) return NULL;
    
    if (!options->chunk_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->chunks) return NULL;
	options->chunk_pool = sctx->chunks;
    }
    
    if (!options->buffer_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->buffers) return NULL;
	options->buffer_pool = sctx->buffers;
    }
    
    dsLockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    
    dm = opcDataCreateContext(ctx->malloc);
    if (dm) {
	err = opcDataInitContext(dm, options);
	if (!err) err = opcDataRegisterNativeRepresentation(dm, type, NULL);
	if (!err) err = opcDataRegisterOPCXMLRepresentation(dm, type, NULL);

	if (err) {
	    opcDataUnref(dm);
	    dm = NULL;
	} else {
	    ctx->dm = (OPCDataManagerContext)dm;
	    ctx->representation = OPC_REPRESENTATION_OPCXML;
	}
    } else err = OPC_ERR_DATA_CREATE;
    
    dsLockReturn(ctx->lock);

    if (err)
	dsDebugServer(DS_DEBUG_WARNING, "opcItem(%p): RegisterLegacyDataManager: %lx (%s)", ctx, err, dsError(err, NULL));
    
    return dm;
}


void opcItemRemoveDataManager(OPCItemContext ctx) {
    OPCDataManagerContext dm;
    ds_assert_return(ctx);

    opcItemDisableDataManager(ctx);

    dsLockWrite(ctx->lock);    
    if (ctx->dm) {
	dm = ctx->dm;
	ctx->dm = NULL;
	opcDataManagerUnref(dm);
    }
    dsLockReturn(ctx->lock);
    
}

void opcItemEnableDataManager(OPCItemContext ctx) {
    dsLockWrite(ctx->lock);
    if (ctx->dm) ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    dsLockReturn(ctx->lock);
}

void opcItemDisableDataManager(OPCItemContext ctx) {
    dsLockWrite(ctx->lock);
    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    dsLockReturn(ctx->lock);
}


OPCDataManagerContext opcItemGetDataManager(OPCItemContext ctx) {
    return NULL;
}

void opcItemReturnDataManager(OPCItemContext ctx) {
}



/*
OPCDataXMLContext opcItemRegisterXMLDataManager(OPCItemContext ctx,
*/

/*
*/
/*
int opcItemPushNativeData(OPCItemContext ctx, DSTime time, OPCQuality quality, void *data) {
    int err;
    ds_assert_param(ctx);

    if (time) {
	dsTimeCopy(&ctx->timestamp, time);
    } else {
	dsTimeCurrentTime(&ctx->timestamp);
	time = &ctx->timestamp;
    }
    
    err=opcDataPushNative(ctx->data, time, quality, data, 0);

    return err;
}

int opcItemGetOPCXMLData(OPCItemContext ctx, OPCDataInfo *info, OPCData *data) {
    ds_assert_param(ctx);
    
    puts("Accquiring Data");
    // ToDo:
//    if (ctx->status&OPC_ITEM_CONNECTED_FLAG)
        return opcDataGetOPCXML(ctx->data, info, data);
    return OPC_ERR_ITEM_NOTCONNECTED;
}
*/
/*
int opcItemGetData(OPCItemContext ctx, unsigned short int format, DSTime time, OPCQuality quality, void *data) {
}

int opcItemGetOPCXMLData(OPCItemContext ctx, DSTime time, OPCQuality quality, void **data) {
    opcDataGet(ctx->data, time, quality, data);
}

int opcItemReturnOPCXMLData(OPCItemContext ctx, void *data) {
}

*/

int opcItemCheckConnection(OPCItemContext ctx) {
    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return OPC_ERR_ITEM_NOTCONNECTED;
    ds_assert_param(ctx->dm);

    return 0;
}

const DSTime opcItemGetTimestamp(OPCItemContext ctx) {
    ds_assert_null(ctx);
    
    return &ctx->timestamp;
}

int opcItemCheckAge(OPCItemContext ctx, const DSTime current_time, const DSDuration maxage) {
    DSTimeS time;

    ds_assert_param(ctx);
    ds_assert_param(maxage);
    
//    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return OPC_ERR_ITEM_NOTCONNECTED;
    
    return 0;
//    if (NOT_APPLICABLE) return 0;
    
    if (!current_time) dsTimeCurrentTime(&time);
    else dsTimeCopy(&time,current_time);
    dsTimeSub(&time,maxage);
    if (dsTimeCmp(&time, &ctx->timestamp)>0) {
//	if (RELOADING_APPLICABLE) OPC_ERR_ITEM_VALUE_OUTOFDATE_REACQUIRING;
	// Ask item to reload if possible otherwise
	
	return OPC_ERR_ITEM_OUTOFDATE;
    }
    return 0;
}


int opcItemGetData(OPCItemContext ctx, OPCDataInfo *info, OPCRepresentationData *data) {
    int err;
    ds_assert_param(ctx);
    
    dsLockRead(ctx->lock);
    if (ctx->status|OPC_ITEM_CONNECTED_FLAG) {
	opcDataManagerRef(ctx->dm);
	dsLockReturn(ctx->lock);
	
	err = opcDataManagerGetData(ctx->dm, ctx->representation, info, data);
	opcDataManagerUnref(ctx->dm);
	return err;
    }
    dsLockReturn(ctx->lock);
    
    return OPC_ERR_ITEM_NOTCONNECTED;
}

void opcItemReturnData(OPCItemContext ctx, OPCDataInfo info) {
    dsLockRead(ctx->lock);
    
    if (ctx->status|OPC_ITEM_CONNECTED_FLAG) {
	opcDataManagerRef(ctx->dm);
	dsLockReturn(ctx->lock);
	opcDataManagerReturnData(ctx->dm, info);
	opcDataManagerUnref(ctx->dm);
    }
    dsLockReturn(ctx->lock);
}

int opcItemPushData(OPCItemContext ctx, OPCDataInfo info, void *data, unsigned long size) {
    // ToDo: Implement
    return 0;
}

// ToDo: GetFirst, GetNext
