#include <dsassert.h>

#include <dslib.h>
#include <opcdatachunk.h>

OPCDataChunkContext opcDataChunkCreateContext(DSMallocContext mctx) {
    int err;
    DSLockContext lock;
    DSConditionContext cond;
    OPCDataChunkContext ctx;
    
    lock = dsLockCreateContext(mctx);
    if (!lock) return NULL;
    err = dsLockInitContext(lock);
    if (err) {
	dsLockDestroyContext(lock);
	return NULL;
    }
    
    cond = dsConditionCreateContext(mctx);
    if (!cond) {
	dsLockDestroyContext(lock);
	return NULL;
    }
    err = dsConditionInitContext(cond);
    if (err) {
	dsLockDestroyContext(lock);
	dsConditionDestroyContext(cond);
	return NULL;
    }

    ctx = (OPCDataChunkContext)dsMalloc(mctx,sizeof(struct OPCDataChunkContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->lock = lock;
	ctx->cond = cond;
	ctx->data = NULL;
	ctx->supported_representations = 0;
    } else {
	dsLockDestroyContext(lock);
	dsConditionDestroyContext(cond);
    }
    return ctx;
}

int opcDataChunkInitContext(OPCDataChunkContext ctx, OPCRepresentationInfo *definfo, unsigned short int supported_representations) {
    long i;
    OPCRepresentationData data;

    ds_assert_param(ctx);
    ds_assert_param(supported_representations);
    
    if ((!ctx->data)||(ctx->supported_representations!=supported_representations)) {
	data = (OPCRepresentationData)dsRealloc(ctx->malloc, ctx->data, supported_representations*sizeof(struct OPCRepresentationDataT));
	if (!data) return DS_ERR_MEM;
	for (i=ctx->supported_representations;i<supported_representations;i++) data[i].data = NULL;
	ctx->data = data;
	ctx->supported_representations=supported_representations;
    }
    return opcDataChunkRestoreContext(ctx, definfo);
}

int opcDataChunkRestoreContext(OPCDataChunkContext ctx, OPCRepresentationInfo *definfo) {
    OPCRepresentationData data;
    long i;

    ctx->next = NULL;
    ctx->referers = 0;
    ctx->delayed_referers = 0;

    data = ctx->data;
    for (i=ctx->supported_representations-1;i>=0;i--) {
	data[i].data = NULL;
	data[i].info = definfo[i];
    }
    return 0;
}


void opcDataChunkFreeContext(OPCDataChunkContext ctx) {
    OPCRepresentationInfo info;
    OPCRepresentationData data;
    short i;
    
    ds_assert_return(ctx);

    data = ctx->data;
    for (i=ctx->supported_representations-1;i>=0;i--) {
        if (data[i].data) {
	    info = data[i].info;
	    switch(info->destructor_type) {
		case OPC_DATA_DESTRUCTOR_RETURN:
		    ((DSReturnFunction)(info->Destructor))(info->ctx,data[i].data);
		break;
		case OPC_DATA_DESTRUCTOR_DESTROY:
		    ((DSDestroyContextFunction)(info->Destructor))(data[i].data);
		break;
		case OPC_DATA_DESTRUCTOR_RETURN_DATA:
		    ((OPCReturnDataFunction)(info->Destructor))(info->ctx,(OPCData)(data+i));
		break;
		case OPC_DATA_DESTRUCTOR_DESTROY_DATA:
		    ((OPCDestroyDataContextFunction)(info->Destructor))((OPCData)(data+i));
		break;
		case OPC_DATA_DESTRUCTOR_FREE:
		    dsFree(info->malloc, data[i].data);
		break;
	    }
	}
	data[i].data = NULL;
    }
}

void opcDataChunkDestroyContext(OPCDataChunkContext ctx) {
    opcDataChunkFreeContext(ctx);
    if (ctx->data) dsFree(ctx->malloc, ctx->data);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->cond) dsConditionDestroyContext(ctx->cond);
    dsFree(ctx->malloc,ctx);
}

int opcDataChunkSet(OPCDataChunkContext ctx, unsigned short int format, void *data, unsigned long size) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->data);
    ds_assert_param(format<ctx->supported_representations);

    ctx->data[format].data = data;
    ctx->data[format].size = size;
    
    return 0;    
}

OPCRepresentationData opcDataChunkGet(OPCDataChunkContext ctx, unsigned short int format) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->data);
    
    if ((format<ctx->supported_representations)&&(ctx->data[format].data)) return ctx->data+format;
    return NULL;
}

OPCRepresentationData opcDataChunkGetData(OPCDataChunkContext ctx, unsigned short int format) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->data);
    
    if (format<ctx->supported_representations)
	return ctx->data+format;

    return NULL;
}

unsigned short opcDataChunkFindData(OPCDataChunkContext ctx) {
    OPCRepresentationData data;
    unsigned short i, representations;

    ds_assert_err(ctx, (unsigned short)-1);
    
    data = ctx->data;
    ds_assert_err(data, (unsigned short)-1);
    
    representations = ctx->supported_representations;
    
    for (i=0;i<representations;i++)
	if (data[i].data) return i;
    return (unsigned short)-1;
}

int opcDataChunkSetInfo(OPCDataChunkContext ctx, const DSTime time, const OPCQuality quality) {
    ds_assert_param(ctx);
    ds_assert_param(time);
    ds_assert_param(quality);
    
    dsTimeCopy(&ctx->info.time, time);
    opcQualityCopy(&ctx->info.quality, quality);
    return 0;
}

int opcDataChunkCopyInfo(OPCDataChunkContext ctx, const OPCDataInfo info) {
    if (info) {
    	if (info->time) dsTimeCopy(&ctx->info.time, info->time);
	else dsTimeCurrentTime(&ctx->info.time);
    
	if (info->quality) opcQualityCopy(&ctx->info.quality, info->quality);
	else opcQualityCopy(&ctx->info.quality,OPCQualityGood);
    } else {
	dsTimeCurrentTime(&ctx->info.time);
	opcQualityCopy(&ctx->info.quality,OPCQualityGood);
    }
    return 0;
}

OPCDataInfo opcDataChunkGetInfo(OPCDataChunkContext ctx) {
    ds_assert_null(ctx);
    
    return (OPCDataInfo)ctx;
}

OPCDataChunkContext opcDataChunkGetChunkByInfo(OPCDataInfo ctx) {
    ds_assert_null(ctx);
    
    return (OPCDataChunkContext)ctx;
}


unsigned long opcDataChunkRef(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkUnref(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkGetRefs(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->referers;
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkRefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkUnrefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkGetRefsByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->referers;
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkDelayedRef(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkRefDelayed(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    (((OPCDataChunkContext)ctx)->delayed_referers)--;
    res=++(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkDelayedUnref(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkGetDelayedRefs(OPCDataChunkContext ctx) {
    unsigned long res;
    dsLockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->delayed_referers;
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkDelayedRefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkRefDelayedByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    (((OPCDataChunkContext)ctx)->delayed_referers)--;
    res=++(((OPCDataChunkContext)ctx)->referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkDelayedUnrefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkGetDelayedRefsByInfo(OPCDataInfo ctx) {
    unsigned long res;
    dsLockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->delayed_referers;
    dsLockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}


void opcDataChunkLock(OPCDataChunkContext ctx) {
    ds_assert_return(ctx);
    dsLockWrite(ctx->lock);
}

void opcDataChunkUnlock(OPCDataChunkContext ctx) {
    ds_assert_return(ctx);
    dsLockReturn(ctx->lock);
}

int opcDataChunkWaitUnref(OPCDataChunkContext ctx, unsigned char delayed) {
    ds_assert_param(ctx);
    if (delayed) {
	dsConditionPreLock(ctx->cond);
	while ((ctx->referers)||(ctx->delayed_referers)) dsConditionWaitLocked(ctx->cond);
	dsConditionUnLock(ctx->cond);
    } else {
	dsConditionPreLock(ctx->cond);
	while (ctx->referers) dsConditionWaitLocked(ctx->cond);
	dsConditionUnLock(ctx->cond);
    }
    return 0;
}

void opcDataChunkRun(OPCDataChunkContext ctx, OPCDataChunkRunFunction func, void *attr) {
    long i;

    ds_assert_return(ctx);

    for (i=ctx->supported_representations-1;i>=0;i--) 
	if (ctx->data[i].data) func(attr, ctx->data+i);
}
